import lodash from 'lodash'
import { basename, sep, relative, join, resolve } from 'path'
import isAbsolutePath from 'path-is-absolute'
import { query, router, mixinPropertyUtils } from 'skypager-util'

import Document from '../document'
import DocumentImporter from '../importer'
import Scope from '../scope'

const { string, set, groupBy, snakeCase, camelCase, uniq, get, at, result, chain, mapValues } = lodash

const { values, keys, defineProperty, assign } = Object

export class Collection {

  /**
   * Create a collection which wraps a directory and creates a queryable index
   * of all of the files inside of it.  Files exist as VFile (@see https://github.com/wooorm/vfile) and are
   * wrapped by Document classes which provide the interface for metadata, transformations, and things of that nature.
   * A collection also provides a lodash wrapper interface for building extraction functions from the collections
   * files and data.
   *
   * @param {Object} options - the options hash
   * @param {String} options.type - the type of document - will determine which document model wrapper is used
   * @param {String} options.base - the root of the collection, defaults to process.cwd()
   * @param {String} options.cwd - the cwd, if different from the base
   * @param {Array} options.ignore - an array of gitignore style patterns of files to ignore
   * @param {Object} options.documentOptions - an object that will be passed to the document wrapper when created
   */
  constructor (options = {}) {
    mixinPropertyUtils(this)

    this.hide('base', options.base || process.cwd())
    this.hide('files', [])
    this.hide('options', { ignore: [], ...options })
    this.hide('root', this.base)

    this.pathId = this.base.split('/').reverse().slice(0,2).join('-')
    this.instanceId = `${this.pathId}-${Math.floor(Date.now() / 1000)}`

    this.importer = DocumentImporter.create(this.base, {
      patterns: this.patterns,
      base: this.base,
      ...(options.importerOptions || {})
    })

    this.hide('cache', {})
    this.hide('_scopes', {})
  }

  static mount (base = process.cwd(), options = {}) {
    if (!isAbsolutePath(base)) {
      base = resolve(process.cwd(), base)
    }

    return create({
      base,
      ...options,
    }).imported.cached
  }

  static create (options = {}) {
    return new Collection(options)
  }

  static get Document() {
    return Document
  }

  static get stringUtils() {
    return require('lodash/string')
  }
  get router() {
    return this.routers.documents
  }

  get routers() {
    if (!this._routers) {
      this.hide('_routers', {})

      this.createRouter('files', {
        pathProperty: 'relative'
      })

      this.createRouter('documents', {
        pathProperty: 'id',
        pathsGetter: 'documents',
      })

      this.createRouter('folders', {
        pathProperty: 'toString',
        pathsGetter: 'subFolderNames'
      })
    }

    return mapValues(this._routers,
      (router) => router.get.bind(router)
    )
  }

  createRouter(name, options = {}) {
    if (options.getter) { options.pathsGetter = options.getter}
    if (options.property) { options.pathProperty = options.property}

    router(this._routers, {
      routerProperty: name,
      pathsGetter: name,
      pathProperty: 'id',
      ...options,
      host: this,
      matchesOnly: true,
    })
  }

  createDelegators(receiver, ...propNames) {
    propNames.forEach(propName => {
      receiver.hideGetter(propName, () => this[propName])
    })
  }

  join(...args) {
    return join(this.root, ...args)
  }

  relative(...args) {
    return relative(this.root, ...args)
  }

  resolve(...args) {
    return resolve(this.root, ...args)
  }

  get(...args) {
    return get(this, ...args)
  }

  result(...args) {
    return result(this, ...args)
  }

  at(...args) {
    return at(this, ...args)
  }

  get chain() {
    return chain(this)
  }

  select(...paths) {
    return this.selectChain(...paths).value()
  }

  selectChain(...paths) {
    const set = (obj,path,val) => {
      lodash.set(obj,path,val)
      return obj
    }

    return this.chain.at(...lodash.sortBy(paths, (p) => p.length))
      .map((value,index) => [paths[index], value])
      .reduce(((memo, pair) => set(memo, ...pair)), {})
  }

  selectByObject (pathObject) {
    const keys = keys(pathObject)
    const paths = values(pathObject)

    return this.chain.at(...paths).reduce(
      ((memo,val,index) => set(memo, keys[index], val)),
      {}
    ).value()
  }

  get filesTree() {
    if (this._filesTree) { return this._filesTree }

    const tree = this.freshFilesTree

    this.hide('_filesTree', tree)

    return tree
  }

  get freshFilesTree () {
    const base = this.subFolderObjectPaths.reduce(
      ((memo, objectPath) => set(memo, objectPath.split('.'), {})), {}
    )

    const filePaths = this.relativeObjectPaths

    this.subFolderObjectPaths.forEach(basePath => {
        const files = get(filePaths, basePath, [])

        files.forEach(d => {
          const baseName = d[0]
          const rel = d[1]
          set(base, basePath.split('.').concat(baseName), this.getFileByRelativePath(rel))
        })
    })

    return base
  }

  get relativeObjectPaths() {
    return this.chain.get('files')
      .sortBy(f => f.dirname.length)
      .map((f) => [
        this.relative(f.dirname).replace(/\//g, '.'),
        f.basename,
        f.relative
      ])
      .groupBy(i => i.shift())
      .value()
  }

  get subFolderObjectPaths() {
    return this.chain.get('subFolderNames')
      .sortBy(v => v.length)
      .map(p => p.replace(/\//g,'.'))
      .value()
  }

  get patterns() {
    return this.typePatterns.concat(this.ignorePatterns)
  }

  get scopes() {
    const collection = this
    const base = {
      available: keys(collection._scopes),
    }

    base.available.forEach(
      scopeName => defineProperty(base, scopeName, {
        configurable: false,
        false: true,
        get: function() {
          const cfg = collection.scope(scopeName)
          return collection.files.filter(file => file.relative.match(cfg.pattern))
        }
      })
    )

    return base
  }

  get ignorePatterns () {
    return this.options.ignore
      .map(p => p.startsWith('!') ? p : `!${p}`)
      .concat([
        // ignore any files contained in the node_modules folder
        '!node_modules/**',
        '!node_modules',
        // ignore any files starting with _
        '!_*',
      ])
  }

  get imported() {
    if(!this.filesLoaded) {
      this.loadFiles(this.importer.readFilesSync(false))
    }

    return this.cached
  }

  get rootFiles() {
    return this.chains.files.filter(f => f.dirname === project.root).value()
  }

  get groupedByChildFolder() {
    return this.chains.files
      .groupBy(f => this.relative(f.dirname).split('/').shift())
      .pickBy((v,k) => k.length > 0)
      .value()
  }

  get groupedByChildFolderAndExtension() {
    return this.chains.files
      .groupBy(f => this.relative(f.dirname).split('/').shift())
      .mapValues(v => groupBy(v, (v) => v.extname))
      .pickBy((v,k) => k.length > 0)
      .value()
  }

  get chains() {
    return {
      documents: chain(this.documents),
      files: chain(this.files),
    }
  }

  get query() {
    const collection = this

    return {
      get files() {
        return query.bind(query, collection.files)
      },
      get documents() {
        return query.bind(query, collection.documents)
      }
    }
  }

  findDocumentById (documentId) {
    return this.documents.find(doc => doc.id === documentId)
  }

  applyFilesRoute(pattern) {
    const path = route()(pattern)
    const results = this.files.map(f => ({path: f.relative, pathData: path(f.relative)})).filter(f => f.pathData)

    return {
      results,
      pattern
    }
  }

  get depths() {
    return this.chains
      .files.map(f => this.relative(f.dirname))
      .uniq()
      .groupBy(f => f.split(sep).length - 1)
      .omit('')
      .value()
  }

  get requireContexts() {
    return chain(this.groupedByChildFolder)
    .mapValues((files, childFolder) => (
      [
        `./${childFolder}`,
        true,
        `\.(${chain(files).map(file => file.extname.replace(/^\./,'')).uniq().value().join('|')})$`
      ]
    )).value()
  }

  get relativeFilePaths() {
    return this.files.map(file => relative(this.root, file.path))
  }

  get relativeSubFolderPaths() {
    return this.files.map(file => relative(this.root, file.dirname))
  }

  get subFolderPaths() {
    return uniq(this.files.map(f => f.dirname))
  }

  get subFolderNames() {
    return this.subFolderPaths
      .map(f => relative(this.root, f))
      .filter(f => f.length > 0)
  }

  get subFolderBasenames () {
    return this.subFolderPaths
      .map(f => relative(this.root, f))
      .map(f => basename(f))
  }

  get childFolderNames() {
    return uniq(
      this.subFolderNames.map(p => p.split('/').shift())
    )
  }

  get childFolderPaths() {
    return uniq(
      this.childFolderNames.map(
        p => this.join(p)
      )
    )
  }

  loadFiles (files) {
    this.files.length = 0

    this.files.push(
      ...files.map(file => {
        return file
      })
    )

    this.filesLoaded = true

    return this.files
  }

  runImporter(cacheDocuments = true) {
    return this.importer.readFiles(false)
      .then(files => this.files.push(...files))
      .then(() => cacheDocuments ? this.cached : this)
      .catch(error => error)
  }

  getFileByRelativePath (relativeFilePath) {
    const {index} = this.filesIndex[relativeFilePath]
    return this.files[index]
  }

  get filesIndex() {
    if (this._filesIndex) { return this._filesIndex }
    this.hide('_filesIndex', this.freshFilesIndex)

    return this._filesIndex
  }

  get freshFilesIndex() {
    return this.chain.get('files')
      .map((f,i) => ({index: i, path: f.path, relative: f.relative}))
      .keyBy(f => f.relative).value()
  }

  get scope() {
    let fn = this.defineScope.bind(this)

    return assign(fn, this.scopes)

    return fn
  }

  defineScope (scopeName, routePattern, options) {
    const name = camelCase(scopeName)

    if (this._scopes[name] && !routePattern && !options) {
      return this._scopes[name]
    }

    return this._scopes[name] = new Scope({
      pattern: typeof routePattern === 'string' ? new RegExp(routePattern) : routePattern,
      key: scopeName,
      name,
      ...(options || {}),
    })
  }

  get ready() {
    return this.importer.status === 'LOADED'
      ? Promise.resolve(this)
      : this.runImporter().then(importer => importer)
  }

  get documentOptions() {
    return {
      base: this.root,
      collectionType: this.options.type,
      cwd: this.options.cwd || this.root,
      ...this.options.documentOptions || {},
    }
  }

  get cached() {
    this.files.forEach(file => {
      this.cache[file.path] = this.cache[file.path]
      || new Document(file.path, this.documentOptions)
    })

    return this
  }

  get documents() {
    return values(this.cache)
  }

  get type() {
    return snakeCase(this.options.type || 'Project').replace(/s$/i, '')
  }

  get typePatterns () {
    switch (this.type.toUpperCase()) {
        case PACKAGE:
          return PatternMapping.Package
          break
        case SCRIPT:
          return PatternMapping.Script
          break
        case SETTINGS_FILE, COPY_FILE:
          return PatternMapping.SettingsFile
          break
        case DOCUMENT:
          return PatternMapping.Document
          break
        case STYLESHEET:
          return PatternMapping.Stylesheet
          break
        case VECTOR:
          return PatternMapping.Vector
          break
        case PROJECT:
        default:
          return PatternMapping.Project
    }
  }
}

export default Collection

const PACKAGE = 'PACKAGE'
const SCRIPT = 'SCRIPT'
const DOCUMENT = 'DOCUMENT'
const SETTINGS_FILE = 'SETTINGS_FILE'
const COPY_FILE = 'COPY_FILE'
const PROJECT = 'PROJECT'
const VECTOR = 'VECTOR'
const STYLESHEET = 'STYLESHEET'

export const PatternMapping = {
  'Package': ['**/package.json'],
  'Document': ['**/*.md'],
  'Script': ['**/*.js','**/*.jsx'],
  'Stylesheet': ['**/*.{css,less,sass,scss,styl'],
  'SettingsFile': ['**/*.yml'],
  'CopyFile': ['**/*.yml'],
  'Vector': ['**/*.svg'],
  'DataSource': ['**/*.json','**/*.yml', '**/*.data.js'],
  'Project': [
    '**/*.{js,jsx,css,styl,less,sass,scss,md,yml,json}',
    '**/package.json',
    '!node_modules/'
  ]
}

export function create(options = {}) {
  return Collection.create(options)
}
